//------------------------------------------------------------------------------
// CERN BE-CO-HT
// Mock Turtle
// https://gitlab.cern.ch/coht/mockturtle
//------------------------------------------------------------------------------
//
// unit name:   MockTurtleDriver
//
// description: A SystemVerilog Class to provide an abstraction of the complete
// MockTurtle core.
//
//------------------------------------------------------------------------------
// Copyright CERN 2018
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`ifndef __MOCK_TURTLE_DRIVER_INCLUDED
 `define __MOCK_TURTLE_DRIVER_INCLUDED

 `include "simdrv_defs.svh"
 `include "if_wishbone_accessor.svh"
 `include "mt_cpu_csr_driver.svh"
 `include "mt_queue_message.svh"
 `include "mt_hmq_driver.svh"
 `include "mt_config_rom_driver.svh"

// Simple interface to deliver the four interrupts of MT to the
// MockTurtleDriver class below
interface IMockTurtleIRQ (
			  input logic hmq_in,
			  input logic hmq_out,
			  input logic notify,
			  input logic console);
endinterface // IMockTurtleIRQ

// Macro to simplify attaching one side of the interface to signals
// in mock_turtle_core.vhd.
// (The "other" side is monitored by the MockTurtleDriver class below)
 `define MT_ATTACH_IRQ(mt)\
.hmq_in  (mt.hmq_in_irq),\
.hmq_out (mt.hmq_out_irq),\
.notify  (mt.notify_irq),\
.console (mt.console_irq)

// A class can only accept virtual interfaces, so we declare it here
typedef virtual IMockTurtleIRQ vIMockTurtleIRQ;

// The main class, to monitor and drive the complete MT
class MockTurtleDriver;
   protected CWishboneAccessor acc;
   protected uint64_t base;
   protected uint32_t core_count;
   protected MTCPUControl csr;
   protected MQueueHost hmq[];
   protected MTConfigRom rom;
   protected vIMockTurtleIRQ irq;

   // When set to 0, parent application will need
   // to clear notification irqs manually
   protected bit autoclear_notify;

   function  new ( CWishboneAccessor acc, uint64_t base,
		   vIMockTurtleIRQ irq, bit autoclear_notify = 1'b1 );
      this.acc  = acc;
      this.base = base;
      this.irq  = irq;
   endfunction // new

   task init ( wb_cycle_type_t mode = PIPELINED );
      uint32_t slot_count;
      wb_set_mode ( mode );
      // first init the config ROM, to be able to get core count etc.
      rom = new ( acc, base + 'he000 );
      rom.init();
      core_count = rom.getCoreCount();
      $display ("App ID: 0x%x", rom.getAppID);
      $display ("Core count: %0d", core_count);
      // Next init the MT CSR
      csr = new ( acc, base + 'hc000, core_count );
      csr.init();
      // Host message queue objects are one per cpu core
      hmq = new [core_count];
      for (int i = 0; i < core_count; i++)
	begin
	   slot_count = rom.getHmqSlotCount(i);
	   // CSR is passed to each queue to be able to select slot
	   // inside the class via csr.hmq_select()
	   hmq[i] = new ( acc, base + 'h0000, i, csr, slot_count );
	   hmq[i].init();
	end
   endtask // init

   task load_firmware ( int core, string filename, bit check = 1'b1);
      csr.load_firmware ( core, filename, check );
   endtask // load_firmware

   task reset_core ( int core, int reset );
      csr.reset_core ( core, reset );
   endtask // reset_core

   task cfg_rom_display();
      rom.pretty_print();
   endtask // cfg_rom_display

   task cfg_rom_dump();
      rom.dump();
   endtask // cfg_rom_dump

   task wb_set_mode ( wb_cycle_type_t mode );
      acc.set_mode ( mode );
   endtask // wb_set_mode

   task enable_console_irq ( int core, int enable );
      csr.uart_int_enable ( core, enable );
   endtask // enable_console_irq

   task enable_hmqi_irq ( int core, int queue, int enable );
      csr.hmq_in_int_enable ( core, queue, enable );
   endtask // enable_hmqi_irq

   task enable_hmqo_irq ( int core, int queue, int enable );
      csr.hmq_out_int_enable ( core, queue, enable );
   endtask // enable_hmqo_irq

   task smem_read ( input uint32_t offset, ref uint32_t val );
      uint64_t tmp;
      acc.read (base + 32'h1_0000 + offset, tmp);
      val = tmp;
   endtask // smem_read

   task smem_write ( uint32_t offset, uint32_t val);
      acc.write (base + 32'h1_0000 + offset, val);
   endtask // smem_write

   task get_cpu_notifications (ref uint32_t val);
      val = 0;
      if (csr.notify_queue.size() == 0)
	return;
      while (csr.notify_queue.size() != 0)
	val |= csr.notify_queue.pop_front();
   endtask // get_cpu_notifications

   task clear_cpu_notifications (uint32_t val);
      csr.clear_notification_irq(val);
   endtask // clear_cpu_notifications

   task handle_hmq_in_irq ();
   endtask // handle_hmq_in_irq

   task hmq_receive_message (ref MQueueMsg msg);
      msg = hmq[msg.core].receive_message (msg.slot);
   endtask // hmq_receive_message

   task hmq_send_message (input MQueueMsg msg);
      hmq[msg.core].send_message (msg.slot, msg);
   endtask // hmq_send_message

   task hmq_purge (int core, int slot);

      hmq[core].purge (slot);
   endtask // hmq_purge

   task hmq_purge_all (int core);
      for (int slot = 0; slot < rom.getHmqSlotCount(core); slot++)
	hmq[core].purge (slot);
   endtask // hmq_purge_all

   function bit hmq_pending_messages (uint32_t core, uint32_t slot);
      return hmq[core].pending_in(slot);
   endfunction // hmq_pending_messages

   task update();
      uint32_t val;
      if (irq.console == 1'b1)
	begin
	   csr.check_consoles();
	end
      if (irq.notify == 1'b1)
	begin
	   csr.handle_notification_irq(autoclear_notify);
	end
      if (irq.hmq_in == 1'b1)
	begin
	   for (int i = 0; i < core_count; i++)
	     begin
		hmq[i].receive_pending();
	     end
	end
      for (int i = 0; i < core_count; i++)
	begin
	   hmq[i].send_pending();
	end
   endtask // update

endclass // MockTurtleDriver

`endif //  `ifndef __MOCK_TURTLE_DRIVER_INCLUDED
